
__module_name__ = "__init__.py"
__doc__ = """plotting __init__ module. Sub-package of the main scdiffeq API."""


# import functions accessed as sdq.pl.<func>: --------------------------------------------
# from .potential_landscape import potential_landscape
from ._velocity_stream import velocity_stream
# from ._smoothed_gene_expression import plot_smoothed_expression as smoothed_expression

# from ._temporal_expression import temporal_expression